from scipy.sparse import csr_matrix, vstack, hstack, issparse

def spblkdiag(A1, A2):
    '''
    SPBLKDIAG  Sparse block diagonal concatenation.

        A = spblkdiag(A1,A2)  

    Given two sparse matrices A1 and A2, this function
    constructs their block diagonal concatenation

        | A1 0  |
    A = |       |
        | 0  A2 |

    where A is a sparse matrix.
    '''

    A = vstack([hstack([A1, csr_matrix((A1.shape[0], A2.shape[1]))]), hstack([csr_matrix((A2.shape[0], A1.shape[1])), A2])])

    if not issparse(A):
        raise Exception('A is not sparse')
    
    return A


import numpy as np
from scipy.linalg import orth
from scipy.spatial import ConvexHull
from scipy.linalg import null_space
import sympy


def useconvhulln(Z2):
    convh = ConvexHull(Z2).simplices
    facets = convh.shape[0]
    nZ1,nZ2 = Z2.shape

    coeff = []
    cold = []
    for i in range(facets):
        vert = np.hstack([Z2[convh[i]],1*np.ones((nZ2,1))])
        M = sympy.Matrix(vert)
        nullvert = np.array(M.nullspace()[0])
        if nullvert.shape[1] == 1:
            if nullvert[-1] != 0:
                nullvert = nullvert / nullvert[-1]
            coeff.append(nullvert[:-1].flatten().tolist())
            cold.extend(nullvert[-1].flatten().tolist())
    coeff = np.array(coeff,dtype='float')
    cold = np.array(cold,dtype='float').reshape(-1,1)

    # Condition the matrix a bit
    coeffcold = np.hstack([coeff,cold])

    # Discard same hyperplanes (due to convhulln result)
    coeff2cold, Ix = np.unique(coeffcold, axis=0, return_index=True)

    # Remove a possible zero row
    yind = [i for i,x in enumerate(np.sum(coeff2cold**2,axis=1)) if x==0]
    if yind:
        Ix = np.hstack([Ix[:yind[0]],Ix[yind[0]+1:]])
        
    coeff2 = coeff[Ix]
    convhnew = convh[Ix]
    cnew = cold[Ix]
    facetsnew = convhnew.shape[0]

    # Make inequalities out of them by testing a point not on the hyperplane
    # Notation: convex hull is now Ax-b<=0
    for fac in range(facetsnew):
        for ind in range(nZ1):
            matr = [i for i,x in enumerate(convhnew[fac]-1) if x==0]
            tests = -1*coeff2[fac] @ Z2[ind].T - cnew[fac]
            if (not matr) and np.abs(tests) > 1e-8:
                break
        if tests > 0:
            coeff2[fac] = -1*coeff2[fac]
            cnew[fac] = -1*cnew[fac]
            
    A = coeff2
    B = cnew

    return A, B



def inconvhull(Z1, Z2):
    # First, find the affine subspace where everything lives
    # (for instance, in the homogeneous case)

    # Translate so it goes throught the origin
    mr = np.mean(Z2, axis=0)
    Rzero = np.array(Z2 - mr)

    # The columns of N generate the subspace
    N = null_space(Rzero)

    # Z2*N should be constant
    cval = np.mean(Z2 @ N, axis=0)

    # Get only the monomials in the subspace
    tol = 0.01
    ix = np.where(np.sum(np.abs(Z1 @ N - cval), axis=1) < tol)
    nZ1 = Z1[ix]

    # Now, the inequalities:
    # Find an orthonormal basis for the subspace
    # (I really should do both things at the same time...)

    # Project to the lower dimensional space, so convhull works nicely
    Q = orth(Rzero.T)

    # This calls CDD, or whatever
    if (Z2 @ Q).shape[1] > 1:
        A, B = useconvhulln(Z2 @ Q)

        # Find the ones tht satisfy the inequalities, and keep them.
        ix = np.where(np.min(B + A@Q.T@Z1[ix].T,axis=0)>-tol)[0]
        Z3 = nZ1[ix]

    elif (Z2 @ Q).shape[1] == 1:
        A = np.array([1,-1]).reshape(-1,1)
        B = np.array([np.max(Z2 @ Q), -np.min(Z2 @ Q)]).reshape(-1,1)
        ix = np.where(np.min(B - A @ Q.T @ nZ1.T, axis=0) > -tol)
        Z3 = nZ1[ix]
    else:
        Z3 = nZ1

    Z3 = np.unique(Z3.toarray(), axis=0)

    return Z3




from sympy import diff, simplify

def sossetobj(sos, symexpr):
    '''
    SOSSETOBJ --- Set the objective function of an SOS program. 

    SOSP = sossetobj(SOSP,EXPR)

    Given a sum of squares program SOSP and an expression EXPR, SOSSETOBJ
    makes EXPR the objective function of the sum of squares program SOSP, 
    i.e., EXPR will be minimized.
    '''

    objvartable_temp = symexpr.free_symbols
    decvartable = sos.symdecvartable

    idx = []
    for i, sym_var in enumerate(objvartable_temp):
        if sym_var not in decvartable:
            raise ValueError('There is unknown decision variable')
        idx.append(decvartable.index(sym_var))
        tempexpr = diff(symexpr, sos.symdecvartable[idx[-1]])
        if len(tempexpr.free_symbols) != 0:
            raise ValueError('Not a linear objective function.')
        else:
            sos.objective = sos.objective.tolil()   # 8/18 added
            sos.objective[idx[-1]] = float(tempexpr)
            sos.objective = sos.objective.tocsr()   # 8/18 added
        symexpr = simplify(symexpr - float(tempexpr)*sos.symdecvartable[idx[-1]])

    return sos
        


from .sosineq import sosineq
from sympy import symbols, Matrix

def sosmatrixineq(sos,fM,option='quadraticMineq'):
    '''
    SOSMATRIXINEQ --- Creates a SOS constraint from a matrix inequality
    constraint

    SOSP = sosmatrixineq(SOSP,fM)

    SOSP is the sum of squares program.
    fM is a polynomial matrix used to generate the polynomial inequality y'fMy>0

    option default is quadraticMineq, it asks for sos expression v'M(x)v
    '''

    [n,m] = fM.shape

    if n != m:
        raise ValueError('Matrix fM in inequality fM>0 must be square.')
    
    if option == 'quadraticMineq':
        # creates the vector of variables Mvar to generate the quadratic expression M_var'*fM*Mvar
        if n > sos.varmat['count']:
            Mvar = list(symbols(f'Mvar_:{n}'))
            Mvar = Mvar[sos.varmat['count']:]
            sos.varmat['symvartable'] = sos.varmat['symvartable'] + Mvar    # updates the vartable in the sos program
            sos.varmat['vartable'] = str(sos.varmat['symvartable'])
            sos.varmat['count'] = n
        
        varMconst = sos.varmat['symvartable'][:n]

        symexpr = Matrix([varMconst]) @ fM @ Matrix([varMconst]).transpose()
        sos = sosineq(sos,symexpr,'sparsemultipartite',[sos.symvartable,varMconst])

    elif option == 'Mineq':
        sos = sosineq(sos,fM)

    return sos






import numpy as np
from scipy.sparse import lil_matrix
import importlib
import cvxpy as cp

class sosprogram:
    def __init__(self, vartable=[], decvartable=[]):
        # Check if SDP solvers are installed
        
        sdp_solvers = ['cvxpy', 'mosek']    # so far only cvxpy and mosek are supported, but cvxpy can call other solvers
        if not any(importlib.util.find_spec(solver) is not None for solver in sdp_solvers):
            raise ImportError('No SDP solvers found.')
        
        SOLVERS = ['MOSEK', 'CVXOPT', 'SCS', 'COPT', 'SDPA']
        INSTALLED_SOLVERS = cp.installed_solvers()
        SOLVER = [item for item in SOLVERS if item in INSTALLED_SOLVERS]
        if len(SOLVER) == 0:
            raise ImportError('No SDP solvers found.')
        print('Installed SDP solvers: ', SOLVER)
        
        # self.var contains information about variables created by initial variables.
        # Z is the matrix of monomial exponents, we can get monomials from Z and vartable.
        # Polynomial or SOS polynomial are indicated by 'type', and their monomials are indicated by 'Z'.
        self.var = {
            'num' : 0,
            'type': {},
            'Z': {},    # monomial degree matrix whose entries are the monomial exponents.
            'ZZ': {},   
            'T': {},
            'idx': {}   # list needs to be initialized, but dict doesn't need to be initialized.
        }

        # self.expr contains information about user's constraints.
        self.expr = {
            'num': 0,
            'type': {},
            'At': {},   # transpose of the coefficient matrix of the constraint
            'b': {},    # right hand side of the constraint
            'Z': {},    # monomial exponents of the constraint
            'multipart': {} # whether the constraint is a multipartite SOS constraint
        }

        # Filled in sossolve(), create decision variables for SDP solver
        self.extravar = {
            'num': 0,
            'Z': {},
            'ZZ': {},
            'T': {},
            'idx': {}
        }

        # weights of the objective function
        self.objective = []     

        # self.solinfo contains information about the SDP solution of the problem.
        self.solinfo = {
            'x': [], 
            'y': [], 
            'RRx': [], 
            'RRy': [], 
            'info': {},
            'solverOptions': {},
            'var' : {},
            'extravar': {},
            'decvar': {}
        }

        # np.shape() reuturns (n,1) for a column vector, but returns (n,) for a row vector.
        # So len(np.shape(vartable)) > 1 indicates that vartable is a column vector.
        # We need to transpose it to a row vector.
        # if vartable is not None and len(np.shape(vartable)) > 1:
        #    vartable = np.transpose(vartable)
        # 7/2 edit: use flatten() to convert whatever shape to a row vector
        if isinstance(vartable, list):
            vartable = np.array(vartable)   # doesn't need flatten()
        else:
            vartable = vartable.flatten()

        if isinstance(decvartable, list):
            decvartable = np.array(decvartable)
        else:
            decvartable = decvartable.flatten()

        self.symvartable = vartable.tolist()        # self.symvartable is a sympy matrix
        self.vartable = str(self.symvartable)       # self.vartable is a string

        # self.varmat contains information about matrix constraints.
        self.varmat = {
            'vartable': '[]', 
            'symvartable': [], 
            'count': 0
        }

        if len(decvartable) != 0:
            self.objective = lil_matrix((len(decvartable), 1))
            self.symdecvartable = decvartable.tolist()      # self.symdecvartable contains decision variables but in a sympy matrix
            self.decvartable = str(self.symdecvartable)     # self.decvartable is a string
            self.var['idx'][0] = len([i for i in self.decvartable if i==',']) + 1   
        else:
            self.decvartable = '[]'
            self.symdecvartable = []
            self.var['idx'][0] = 0


import numpy as np
import math
from scipy.sparse import csr_matrix, lil_matrix, vstack, eye, kron, hstack, issparse, csc_matrix, coo_matrix
import importlib

from .monomials import monomials
from .inconvhull import inconvhull
from .getconstraint import getconstraint
from .findcommonZ import findcommonZ
from .spantiblkdiag import spantiblkdiag
from .spblkdiag import spblkdiag
from .sparsemultipart import sparsemultipart
from .sosmoseksolver import sosmoseksolver
from .soscvxoptsolver import soscvxoptsolver
from .soscvxpysolver import soscvxpysolver


def sparse_fliplr(A):
    A_coo = A.tocoo()
    flipped_col = A.shape[1] - A_coo.col - 1
    flipped_A = coo_matrix((A_coo.data, (A_coo.row, flipped_col)), shape=A_coo.shape)
    return flipped_A.tocsr()

def sparse_flipud(A):
    A_coo = A.tocoo()
    flipped_row = A.shape[0] - A_coo.row - 1
    flipped_A = coo_matrix((A_coo.data, (flipped_row, A_coo.col)), shape=A_coo.shape)
    return flipped_A.tocsr()

def addextrasosvar(sos, I):
    feasextrasos = 1
    for i in I:
        numstates = sos.expr['Z'][i].shape[1]
        maxdeg = int(np.max(np.sum(sos.expr['Z'][i], axis=1)))  # maximum total degree of the monomials
        mindeg = int(np.min(np.sum(sos.expr['Z'][i], axis=1)))  # minimum total degree of the monomials
        Z = monomials(numstates, range(math.floor(mindeg/2), math.ceil(maxdeg/2) + 1))

        # Discarding unnecessary monomials
        maxdegree = np.max(sos.expr['Z'][i], axis=0) / 2    # since sos.expr['Z'][i] is a sparse matrix, maxdegree is a 1D array
        mindegree = np.min(sos.expr['Z'][i], axis=0) / 2    # since sos.expr['Z'][i] is a sparse matrix, mindegree is a 1D array

        Zdummy1 = maxdegree.toarray() - Z.toarray()   # use toarray() to convert a sparse matrix to a dense matrix so that we can do element-wise subtraction
        Zdummy2 = Z.toarray() - mindegree.toarray()
        I, _ = np.where(np.concatenate((Zdummy1, Zdummy2), axis=1) < 0) # find the rows with negative entries
        IND =  np.setdiff1d(np.arange(Z.shape[0]), I)   # find the rows with non-negative entries
        if len(IND) == 0:
            print("\n Warning: Inequality constraint {} in your sos program structure corresponds to a polynomial that is not sum-of-squares.\n".format(i))
            feasextrasos = 0
            return
        else:
            Z = Z[IND, :]   # Only keep the rows with non-negative entries


        if sos.expr['type'][i] == 'sparse':
           Z2 = sos.expr['Z'][i]/2
           Z = inconvhull(Z,Z2)
           Z = csr_matrix(Z)
        
        if sos.expr['type'][i] == 'sparsemultipartite':
            Z2 = sos.expr['Z'][i]/2
            info2 = sos.expr['multipart'][i]
            sizeinfo2m = len(info2)
            vecindex = []
            for indm in range(sizeinfo2m):
                vecindex.append([])
                for indn in range(len(info2[indm])):
                    varcheckindex = [i for i, y in enumerate(sos.symvartable) if info2[indm][indn] == y] # Assume info2[indm][indn] is a symbol
                    if varcheckindex:
                        vecindex[indm].extend(varcheckindex)
                    else:
                        vecindex[indm].extend([len(info2[0]) + [i for i, y in enumerate(sos.varmat['symvartable']) if info2[indm][indn] == y][0]])
            
            Z = sparsemultipart(Z,Z2,vecindex)
            Z = csr_matrix(Z)
        # Z in the quadratic representation has now been updated.

        dimp = sos.expr['b'][i].shape[1]   # detecting whether Mineq is active

        # Adding slack variables
        sos.extravar['num'] += 1
        var = sos.extravar['num'] - 1
        sos.extravar['Z'][var] = csr_matrix(Z)
        [T,ZZ] = getconstraint(Z)
        sos.extravar['ZZ'][var] = csr_matrix(ZZ)
        sos.extravar['T'][var] = csr_matrix(T.T)

        sos.extravar['idx'][var+1] = sos.extravar['idx'][var] + (Z.shape[0] * dimp) ** 2
        for j in range(sos.expr['num']):
            sos.expr['At'][j] = vstack([sos.expr['At'][j], csr_matrix((sos.extravar['T'][var].shape[0]*dimp**2, sos.expr['At'][j].shape[1]))])


        if dimp == 1:
            pc = {}
            pc['Z'] = sos.extravar['ZZ'][var]
            pc['F'] = -eye(pc['Z'].shape[0])
            R1, R2, newZ = findcommonZ(sos.expr['Z'][i], pc['Z'])

            if sos.expr['At'][i].size == 0:
                sos.expr['At'][i] = csr_matrix((sos.expr['At'][i].shape[0], R1.shape[0]))
            
            sos.expr['At'][i] = (sos.expr['At'][i] @ R1).tolil()
            lidx = sos.extravar['idx'][var]
            uidx = sos.extravar['idx'][var+1]
            sos.expr['At'][i][lidx:uidx,:] = sos.expr['At'][i][lidx:uidx,:] - sos.extravar['T'][var] @ pc['F'] @ R2
            sos.expr['At'][i] = sos.expr['At'][i].tocsr()
            sos.expr['b'][i] = R1.T @ sos.expr['b'][i]
            sos.expr['Z'][i] = newZ

        else:
            ZZ = sparse_flipud(ZZ)
            T = sparse_flipud(T)
            Zcheck = sos.expr['Z'][i]
            R1, R2, Znew = findcommonZ(Zcheck, ZZ)

            R1 = sparse_fliplr(R1)
            R2 = sparse_fliplr(R2)
            Znew = sparse_flipud(Znew)

            R1sum = R1.sum(axis=0).tolist()[0]
            T = R2.T @ T

            ii = 0
            #sig_ZZ = ZZ.shape[0]
            sig_Z = Z.shape[0]
            sig_Znew = Znew.shape[0]

            Tf = lil_matrix((dimp**2 * sig_Znew, (dimp * sig_Z)**2))
            Sv = lil_matrix((sig_Znew * dimp**2, 1))
            for j in range(sig_Znew):
                Mt0 = lil_matrix((dimp, dimp * sig_Z**2))
                for k in range(sig_Z):
                    Mt0[:, (dimp*sig_Z)*k: (dimp*sig_Z)*(k+1)] = kron(eye(dimp), T[j, sig_Z*k: sig_Z*(k+1)])
                Tf[j*dimp**2:(j+1)*dimp**2, :] = kron(eye(dimp), Mt0)

                if R1sum[j] == 1:
                    b_temp = sos.expr['b'][i].tolil()
                    Sv[j*dimp**2:(j+1)*dimp**2] = np.reshape(b_temp[dimp*ii:dimp*(ii+1), :].T, (1, dimp**2)).T
                    if ii < Zcheck.shape[0]-1:
                        ii += 1
                else:
                    # sparse matrix slicing, need to be careful. probably need to convert to csc matrix
                    sos.expr['At'][i] = hstack([sos.expr['At'][i][:,:j*dimp**2], csr_matrix((sos.expr['At'][i].shape[0], dimp**2)), sos.expr['At'][i][:,j*dimp**2:]])

            lidx = sos.extravar['idx'][var]
            uidx = sos.extravar['idx'][var+1]
            sos.expr['At'][i] = sos.expr['At'][i].tolil()
            sos.expr['At'][i][lidx:uidx,:] = Tf.tocsr().transpose()
            sos.expr['At'][i] = sos.expr['At'][i].tocsr()
            sos.expr['b'][i] = Sv.tocsr()

    return sos, feasextrasos


def addextrasosvar2(sos, I):
    numstates = sos.expr['Z'][0].shape[1]
    for i in I:
        # Create extra variables
        maxdeg = int(np.max(np.sum(sos.expr['Z'][i], axis=1)))
        mindeg = int(np.min(np.sum(sos.expr['Z'][i], axis=1)))
        Z = monomials(numstates, range(math.floor(mindeg/2), math.ceil(maxdeg/2) + 1))

        # Discarding unnecessary monomials
        maxdegree = sos.expr['Z'][i].max(axis=0) / 2
        mindegree = sos.expr['Z'][i].min(axis=0) / 2
        j = 0
        while j < Z.shape[0]:
            Zdummy1 = maxdegree.toarray() - Z[j].toarray()
            Zdummy2 = Z[j].toarray() - mindegree.toarray()
            idx, _ = np.where(np.concatenate((Zdummy1, Zdummy2), axis=1) < 0)
            if idx.size > 0:
                Z = np.vstack([Z[:j], Z[j+1:]])
            else:
                j += 1
            
        # Add the variables
        for k in range(2):
            sos.extravar['num'] += 1
            var = sos.extravar['num'] - 1
            sos.extravar['Z'][var] = csr_matrix(Z)
            T,ZZ = getconstraint(Z)
            sos.extravar['ZZ'][var] = ZZ    
            sos.extravar['T'][var] = T.T
            sos.extravar['idx'][var+1] = sos.extravar['idx'][var] + Z.shape[0]**2
            for j in range(sos.expr['num']):
                sos.expr['At'][j] = vstack([sos.expr['At'][j], csr_matrix((sos.extravar['T'][var].shape[0], sos.expr['At'][j].shape[1]))])

            # Modifying expression
            degoffset = hstack([k, csr_matrix((1,numstates-1))])
            pc = {}
            pc['Z'] = sos.extravar['ZZ'][var] + vstack([degoffset]*sos.extravar['ZZ'][var].shape[0])
            pc['F'] = -eye(pc['Z'].shape[0])
            R1, R2, newZ = findcommonZ(sos.expr['Z'][i], pc['Z'])

            if sos.expr['At'][i].size == 0:
                sos.expr['At'][i] = csr_matrix((sos.expr['At'][i].shape[0], R1.shape[0]))

            sos.expr['At'][i] = sos.expr['At'][i] @ R1
            lidx = sos.extravar['idx'][var]
            uidx = sos.extravar['idx'][var+1]
            sos.expr['At'][i] = sos.expr['At'][i].tolil()
            sos.expr['At'][i][lidx:uidx,:] = sos.expr['At'][i][lidx:uidx,:] - sos.extravar['T'][var] @ pc['F'] @ R2
            sos.expr['At'][i] = sos.expr['At'][i].tocsr()
            sos.expr['b'][i] = R1.T @ sos.expr['b'][i]
            sos.expr['Z'][i] = newZ

            Z = csr_matrix(Z[Z.toarray()<maxdegree.data[0]].reshape(-1,1))

    return sos


def processvars(sos, Atf, bf):
    # Processing all variables

    # Decision variables
    K = {}
    K['s'] = []
    K['f'] = sos.var['idx'][0]
    RR = eye(K['f'])

    # Polynomial and SOS variables
    for i in range(sos.var['num']):
        if sos.var['type'][i] == 'poly':
            sizeX = int(sos.var['idx'][i+1] - sos.var['idx'][i])
            RR = spantiblkdiag(RR, eye(sizeX))
            K['f'] = sizeX + K['f']
        elif sos.var['type'][i] == 'sos':
            sizeX = int(math.sqrt(sos.var['idx'][i+1] - sos.var['idx'][i]))
            RR = spblkdiag(RR, eye(sizeX**2))
            K['s'].append(sizeX)

    for i in range(sos.extravar['num']):
        sizeX = int(math.sqrt(sos.extravar['idx'][i+1] - sos.extravar['idx'][i]))
        RR = spblkdiag(RR, eye(sizeX**2))
        K['s'].append(sizeX)

    At = RR.T @ Atf

    b = bf

    return At, b, K, RR



def sossolve(sos,options=None,verbose=1):
    '''
    SOSSOLVE --- Solve a sum of squares program.

    SOSP = sossolve(SOSP)

    SOSP is the SOS program to be solved.

    Alternatively, SOSP = sossolve(SOSP,SOLVER_OPT) also defines the solver
    and/or the solver-specific options respectively by fields

    SOLVER_OPT.solver (name of the solver). This can be 'cvxopt,mosek,scs,copt,sdpa'.
    SOLVER_OPT.params (a structure containing solver-specific parameters)

    The default values for solvers is 'cvxopt'.
    '''
    SOLVERS = ['MOSEK', 'CVXOPT', 'SCS', 'COPT', 'SDPA']
    own_mosek=0     # check if you have mosek license
    if importlib.util.find_spec('mosek') is not None:
        own_mosek=1

    # Default options
    if options is None:
        options = {}
        options['solver'] = 'cvxopt'
        # params = {}
        # params['tol'] = 1e-9
        # params['alg'] = 2
        # options['params'] = params
    else:
        if not 'solver' in options:
            options['solver'] = 'cvxopt'
        # if not 'params' in options:
        #     params = {}
        #     params['tol'] = 1e-9
        #     params['alg'] = 2
        #     options['params'] = params

    # # Check sospsimplify
    # if 'simplify' in options:
    #     temp_simp = options['simplify']
    #     if temp_simp == 'on' or temp_simp == 1 or temp_simp.lower() == '1' or temp_simp == 'turn on' or temp_simp == 'simplify':
    #         options['simplify'] = 1
    #     elif temp_simp == 'off' or temp_simp == 0 or temp_simp.lower() == '0' or temp_simp == 'turn off':
    #         options['simplify'] =0
    #     else:
    #         warnings.warn("options.simplify should be one of the options 'on', 1, 'simplify', 'off', '0'")
    #         warnings.warn("Set simplify off by default")
    #         options['simplify'] = 0
    # else:
    #     # Defult simplify is off
    #     options['simplify'] = 0

    # Check if the SOS program is already solved
    if len(sos.solinfo['x']) != 0:
        raise ValueError('The SOS program is already solved.')
    
    # Adding slack variables to inequalities
    sos.extravar['idx'][0] = sos.var['idx'][sos.var['num']]
    # SOS variables
    feasextrasos = 1   # Will be set to 0 if constraints are immediately found infeasible
    I = [i for i, t in sos.expr['type'].items() if t in {'ineq','sparse','sparsemultipartite'}]
    if I:
        sos, feasextrasos = addextrasosvar(sos, I)
    # SOS variables type II (restrictedd on interval)
    I = [i for i, t in sos.expr['type'].items() if t == 'posint']
    if I:
        sos = addextrasosvar2(sos, I)

    # Processing all expressions
    Atf = hstack([sos.expr['At'][i] for i in range(sos.expr['num'])])
    bf = vstack([sos.expr['b'][i] for i in range(sos.expr['num'])])

    # Processing all variables
    At, b, K, RR = processvars(sos, Atf, bf)

    # Objective function
    c = lil_matrix((At.shape[0],1))

    indend = len(sos.var['idx'])
    indend_value = sos.var['idx'][indend-1]
    if issparse(sos.objective):
        objdim = sos.objective.shape[0]
    else:
        objdim = len(sos.objective)
    if objdim == 0:
        sos.objective = np.zeros(c[:indend_value].shape)
    
    c[:indend_value] = c[:indend_value] + sos.objective
    c = RR.T @ c

    #pars = options['params']

    feassosp = 1

    ############currently ignore those options############
    # if options['simplify']==1:  
        
    #     print('Running Simplification process:')
    #     At_full = At.T
    #     c_full = c
    #     b_full = b
    #     K_full = K
    #     size_At_full = At_full.shape
    #     # Some initial parameters for sospsimplify
    #     Nsosvarc = len(K['s'])
    #     dv2x = np.array([i for i in range(size_At_full[1])])
    #     K_full['l'] = 0

    #     Zmonom = {}
    #     for i in range(Nsosvarc):
    #         Zmonom['i'] = np.array([i for i in range(K['s'][i])]).T
    #     if 'tol' not in pars:
    #         ptol = 1e-9
    #     else:
    #         ptol = pars['tol']

    #     # A,b,K, reduced matrices
    #     A,b,K,z,dv2x,Nfv,feassosp,zrem,removed_rows = sospsimplify(At_full,b_full,K_full,Zmonom,dv2x,Nsosvarc,ptol)

    #     print('Old A size: ', size_At_full.T)
    #     # continue at line 189 in MATLAB code
    ######################################################

    # if the addextrasosvar return infeasible solution.
    # If the problem is clearly infeasible, sedumi can return error
    # Return no solution if the problem is clearly infeasible from
    # addextrasosvar.
    if feassosp==0 or feasextrasos==0:
        print('Warning: Primal program infeasible, no solution produced.')
        info = {}
        info['iter'] = 0
        info['feasratio'] = -1
        info['pinf'] = 1
        info['dinf'] = 1
        #info['numerr'] = 0
        #info['timing'] = 0
        info['cpusec'] = 0
        sos.solinfo['info'] = info
        sos.solinfo['solverOptions'] = options
        return sos
    
    ##################################### Solve the problem #####################################
    SOLVER = None
    if options['solver'].lower() == 'mosek' and own_mosek == 1:
        x, y, info = sosmoseksolver(At, b, c, K, verbose)
    elif options['solver'].lower() == 'mosek' and own_mosek == 0:
        print('You do not have mosek license, trying to use mosek by cvxpy solver...')
        SOLVER = 'MOSEK'
    elif options['solver'].lower() == 'cvxopt':
        SOLVER = 'CVXOPT'
    elif options['solver'].lower() == 'scs':
        SOLVER = 'SCS'
    elif options['solver'].lower() == 'copt':
        SOLVER = 'COPT'
    elif options['solver'].lower() == 'sdpa':
        SOLVER = 'SDPA'
    

    ##################################### try different solvers #####################################
    if SOLVER is not None:
        ALTER_SOLVER = [item for item in SOLVERS if item != SOLVER]
        try:
            x, y, info = soscvxpysolver(At, b, c, K, verbose, SOLVER=SOLVER)
        except:
            print(SOLVER,' solver failed, trying ', ALTER_SOLVER[0], ' solver...')
            for i in range(4):
                try:
                    x, y, info = soscvxpysolver(At, b, c, K, verbose, SOLVER=ALTER_SOLVER[i])
                    break
                except:
                    if i == 3:
                        print('All solvers failed.')
                    else:
                        print(ALTER_SOLVER[i],' solver failed, trying ', ALTER_SOLVER[i+1], ' solver...')
    


    ################################### record the solution ###################################
    sos.solinfo['x'] = x
    sos.solinfo['y'] = y

    # 
    if options['solver'] == 'mosek':
        if not x and info['pinf']:
            print('Warning: Primal problem infeasible, no solution produced.')
            info['iter'] = 0
            info['feasratio'] = -1
            info['pinf'] = 1
            info['dinf'] = 1
            #info['numerr'] = 0
            #info['timing'] = 0
            info['cpusec'] = 0
            sos.solinfo['info'] = info
            sos.solinfo['solverOptions'] = options
            return sos
    elif options['solver'] == 'cvxopt':
        if info['status'] == 'infeasible':
            print('Warning: Problem infeasible, no solution produced.')
            info['iter'] = 0
            info['feasratio'] = -1
            info['pinf'] = 1
            info['dinf'] = 1
            info['cpusec'] = 0
            sos.solinfo['info'] = info
            sos.solinfo['solverOptions'] = options
            return sos
    
    # simplify and frlib
    # if options['simplify'] == 1:
    # elif 

    print('\n','Residual norm', np.linalg.norm((At.T.dot(csr_matrix(x).T)-b).toarray().flatten()),'\n')

    sos.solinfo['RRx'] = (RR @ csc_matrix(x).T).toarray().flatten().tolist()
    sos.solinfo['RRy'] = (RR @ (c - At @ csc_matrix(y).T)).toarray().flatten().tolist()
    sos.solinfo['info'] = info
    sos.solinfo['solverOptions'] = options

    # print info
    for key, value in info.items():
        print(f'{key}: {value}')
    print('\n')

    # Constructing the (primal and dual) solution vectors and matrices
    # If you want to have them, comment/delete the return command above.
    # In the future version, these primal and dual solutions will be computed only
    # when they are needed. We don't want to store redundant info.
    for i in range(sos.var['num']):
        if sos.var['type'] == 'poly':
            var = {}
            var['primal'][i] = sos.solinfo['RRx'][sos.var['idx'][i]:sos.var['idx'][i+1]]
            var['dual'][i] = sos.solinfo['RRy'][sos.var['idx'][i]:sos.var['idx'][i+1]]
            sos.solinfo['var'] = var
        elif sos.var['type'] == 'sos':
            var = {}
            primal = {}
            dual = {}
            primaltemp = sos.solinfo['RRx'][sos.var['idx'][i]:sos.var['idx'][i+1]]
            dualtemp = sos.solinfo['RRy'][sos.var['idx'][i]:sos.var['idx'][i+1]]
            primal[i] = np.array(primaltemp).reshape(int(math.sqrt(len(primaltemp))),int(math.sqrt(len(primaltemp)))).T
            dual[i] = np.array(dualtemp).reshape(int(math.sqrt(len(dualtemp))),int(math.sqrt(len(dualtemp)))).T
            var['primal'] = primal
            var['dual'] = dual
            sos.solinfo['var'] = var
    
    for i in range(sos.extravar['num']):
        extravar = {}
        primal = {}
        dual = {}
        primaltemp = sos.solinfo['RRx'][sos.extravar['idx'][i]:sos.extravar['idx'][i+1]]
        dualtemp = sos.solinfo['RRy'][sos.extravar['idx'][i]:sos.extravar['idx'][i+1]]
        primal[i] = np.array(primaltemp).reshape(int(math.sqrt(len(primaltemp))),int(math.sqrt(len(primaltemp)))).T
        dual[i] = np.array(dualtemp).reshape(int(math.sqrt(len(dualtemp))),int(math.sqrt(len(dualtemp)))).T
        extravar['primal'] = primal
        extravar['dual'] = dual
        sos.solinfo['extravar'] = extravar

    decvar = {}
    decvar['primal'] = sos.solinfo['RRx'][:sos.var['idx'][0]]
    decvar['dual'] = sos.solinfo['RRy'][:sos.var['idx'][0]]
    sos.solinfo['decvar'] = decvar

    return sos



import cvxpy as cp
import numpy as np
import time


def soscvxpysolver(At,b,c,K,verbose=1,SOLVER='CVXOPT'):
    '''
    This function is to solve the optimization problem with MOSEK solver

    Output:
        x[list]: primal problem solution
        y[list]: dual problem solution
        info[dic]: information about the solution
    '''
    start_time = time.time()
    if 'q' in K:
        raise Exception('K.q and K.r are supposed to be empty')

    if 'f' not in K:
        K['f'] = []
    if 'l' not in K:
        K['l'] = []

    nf = int(np.sum(K['f']))
    nl = int(np.sum(K['l']))

    Ks2 = [int(item**2) for item in K['s']]
    Kindex = nf + nl + np.cumsum([0]+Ks2)
    RSPind = Ks2 + [nf]

    nsdpvar = len(K['s'])   # Number of Semidefinite Program Variables
    nvars = Kindex[0]       # Number of scalar variable

    A_DIC = []
    C_DIC = []
    VAR_DIC = []
    constraints = []

    ######################### Create Semidefinite variables #########################
    for i in range(nsdpvar):
        VAR = cp.Variable((K['s'][i],K['s'][i]), PSD=True)    # Create a PSD variable
        #constraints += [VAR >> 0]
        VAR_DIC.append(VAR)    # Add to the dictionary
        A = []
        A_temp = At[Kindex[i]:Kindex[i+1]].T
        for j in range(A_temp.shape[0]):
            A.append(cp.trace(A_temp[j].toarray().reshape(K['s'][i],K['s'][i]) @ VAR))  # Trace of matrix multiplication
        A_DIC.append(A)
        C_temp = c[Kindex[i]:Kindex[i+1]]
        C_DIC.append(cp.trace(C_temp.toarray().reshape(K['s'][i],K['s'][i]) @ VAR))  # Trace of matrix multiplication
        
    ################### Create scalar variables, add to the end of VAR_DIC ###################
    if nvars > 0:
        VAR = cp.Variable(nvars)    # Create scalar variables
        VAR_DIC.append(VAR)
        A = []
        A_temp = At[:nvars].T
        for i in range(A_temp.shape[0]):
            A.append(A_temp[i].toarray() @ VAR)
        A_DIC.append(A)
        C_temp = c[:nvars]
        C_DIC.append(C_temp.toarray().T @ VAR)
        
    ########################### Contruct objective function ############################
    OBJ_SUM = sum(C_DIC)

    ############################# Construct constriants ################################
    B = b.toarray().flatten().tolist()
    for i in range(b.shape[0]):
        CON = 0
        for j in range(len(A_DIC)):
            CON += A_DIC[j][i]
        constraints += [CON == B[i]]
        
    ########################### Solve the problem ######################################
    prob = cp.Problem(cp.Minimize(OBJ_SUM),constraints)
    if SOLVER == 'CVXOPT':
        prob.solve(solver = cp.CVXOPT, verbose=verbose)
    elif SOLVER == 'MOSEK':
        prob.solve(solver = cp.MOSEK, verbose=verbose)
    elif SOLVER == 'SCS':
        prob.solve(solver = cp.SCS, verbose=verbose)
    elif SOLVER == 'COPT':
        prob.solve(solver = cp.COPT, verbose=verbose)
    elif SOLVER == 'SDPA':
        prob.solve(solver = cp.SDPA, verbose=verbose)
    
    end_time = time.time()

    ################################# Process solution #################################
    # primal solution
    x = []
    if nvars > 0:
        x.extend(VAR_DIC[-1].value.tolist())
    for i in range(nsdpvar):
        x.extend(VAR_DIC[i].value.flatten().tolist())

    # dual solution
    y = []
    for i in range(len(B)):
        dual_value = constraints[i].dual_value
        if isinstance(dual_value, (float,int,np.float64)):
            y.append(dual_value)
        else:
            y.extend(constraints[i].dual_value)

    # calculate feasratio
    violations = 0
    for constraint in constraints:
        residual = constraint.residual
        violations += np.sum(residual[residual > 1e-5])

    feasratio = 1 - violations / len(constraints)

    # info
    info = {}
    info['cpusec'] = round(end_time-start_time,5)
    info['iter'] = prob.solver_stats.num_iters
    info['feasratio'] = feasratio
    info['status'] = prob.status

    return x,y,info



    



import numpy as np
from fractions import Fraction

def proj3(x0,A,b,N):
    '''
    Orthogonal projection of the point x0 onto the subspace Ax=b 

    Here, all rational data
    A is assumed fat, and full row rank
    A*A' is diagonal

    All integer computations
    The solution is x/NN, the input is x0/N
    '''
    # x0, A, b are all numpy arrays
    # N is an integer
    
    aa = np.sum(A * A, axis=1)
    N2 = np.lcm.reduce(np.unique(aa).astype(int).tolist())
    x = N2*x0 - A.T @ ((N2 / aa).reshape(-1,1) * (A@x0-b*N))

    NN = N2*N

    Den = [[Fraction(val).limit_denominator().denominator for val in row] for row in x]
    x = x * np.prod(np.unique(Den))
    NN = NN * np.prod(np.unique(Den))

    # Make sure it's minimal
    ix = np.unique(abs(x)).nonzero()
    try:
        cfact = np.gcd.reduce(np.unique(abs(x))[ix].astype(int).tolist() + [NN])
    except:
        raise Exception('No reasonable projection with rational values is possible')
        
    x = x / cfact
    NN = NN / cfact

    return x,NN






from scipy.sparse import csr_matrix, vstack, eye, issparse, lil_matrix
from .findcommonZ import findcommonZ

def getconstraint(Z):
    '''
    GETCONSTRAINT --- Find constraint for sum of squares decomposition.

    A,ZZ = getconstraint(Z)

    Z is a monomial vector description.
    This function computes the constraint matrix A and the polynomial
    vector ZZ, such that if q satisfies

    A*q = F'

    Then

    Z'*Q*Z = F*ZZ      (where Q is the matrix form of q)
    '''

    # First write Z'*Q*Z as
    # Z'*Q*Z = (e1'*Q*R1 + e2'*Q*R2 + ... + en'*Q*Rn) ZZ

    if not issparse(Z):
        Z = csr_matrix(Z)

    sizeZ = Z.shape[0]
    ZZ = Z + vstack([Z[0]]*sizeZ)
    M = {}  # Let M be a dictionary, M(1).R = M{1} ...
    M[0] = eye(sizeZ)
    for i in range(1,sizeZ):
        Ztemp = Z + vstack([Z[i]]*sizeZ)
        R1,R2,ZZ = findcommonZ(ZZ,Ztemp)
        for j in range(i):
            M[j] = M[j] @ R1
        M[i] = R2

    # Construc the constaint equations
    Q = lil_matrix((sizeZ,sizeZ))
    A = lil_matrix((ZZ.shape[0],sizeZ**2))
    for i in range(sizeZ**2):
        Q[i % sizeZ, i // sizeZ] = 1
        j, k = Q.nonzero()
        A[:,i] = M[j[0]].T @ Q[j[0],:].T
        Q[i % sizeZ, i // sizeZ] = 0

    A = csr_matrix(A)
    return A, ZZ





import numpy as np
from itertools import combinations_with_replacement
from scipy.sparse import csr_matrix
import sympy



def monomials(vartable,degree):
    '''
    Construct a vector of monomials with prespecified degrees.
    Z = monomials(VARTABLE, DEGREE)

    Given a list of independent variables VARTABLE and a list of non-negative 
    integers DEGREE, this function constructs a column vector 
    containing all possible monomials of degree described in DEGREE.

    For example, monomials([x1,x2],[1,2,3]) with x1 and x2 
    being symbolic variables will return exponent matrix of monomials in x1 and 
    x2 of degree 1, 2, and 3.

    
    Alternative calling convention:

    Z = monomials(n,d)

    Given two integers n and d, this function constructs 
    a vector containing all possible monomials of degree d 
    in n variables.

    n is the number of variables
    d is the degree of monomials
    Z is the exponent matrix of monomial vector
    '''

    # Check if vartable is a single integer, if so consider it as the number of variables
    if isinstance(vartable, int):
        num_variables = vartable
    else:
        num_variables = len(vartable)
    
    Z = []

    for d in degree:
        for comb in combinations_with_replacement(range(num_variables), d):
            z = np.zeros(num_variables, dtype=int)
            for idx in comb:
                z[idx] += 1
            Z.append(z)
    
    if isinstance(vartable, list) and all(isinstance(item, sympy.core.symbol.Symbol) for item in vartable):
        res = []
        for row in Z:
            exprs = [var**deg for var, deg in zip(vartable, row)]
            res.append(sympy.prod(exprs))
        return res
    else:
        return csr_matrix(Z)


import numpy as np
import pandas as pd
from sympy import expand, collect, poly, lambdify, Matrix, together, cancel, Add
from scipy.sparse import csr_matrix, lil_matrix
from sympy.tensor.array import derive_by_array
import sympy

def sortNoRepeat(Z, Zin):
    if Zin.size == 0:
        return Z
    if Z.size == 0:
        return pd.DataFrame(Zin).sort_values(by=list(range(Zin.shape[1])),ascending=False).values
    
    Z = pd.DataFrame(Z)
    Zin = pd.DataFrame(Zin)
    Znew = pd.concat([Z, Zin]).drop_duplicates().values

    Znew = pd.DataFrame(Znew).sort_values(by=list(range(Znew.shape[1])),ascending=False).values

    return Znew


def approx_zero(expr, tolerance=1e-10):
    new_terms = []
    for term in expr.as_ordered_terms():
        coefficient, _ = term.as_coeff_Mul()
        if abs(coefficient) > tolerance:
            new_terms.append(term)
    return Add(*new_terms)


def poly_approx_zero(polynomial):
    '''
    Round small coefficients to zero in a polynomial.
    '''
    numerator, denominator = polynomial.as_numer_denom()
    new_numerator = approx_zero(numerator)
    new_denominator = approx_zero(denominator)

    new_polynomial = new_numerator / new_denominator

    return new_polynomial


def getequation(expr, symvartable, decvartable, varmat):
    '''
    function At,b,Z = getequation(expr,symvartable,decvartable,decvartablename)

    GETEQUATION --- Convert a symbolic expression to At, b, and Z
            used in an SOS program. In this format, 
    expr = Im kron [1][ C ]In kron Z
                   [x][   ]                                                                
         = (b+x^TAt^T) Imn kron Z
    Inputs:
    expr: This expression is in symbolic format. 
    It may be matrix or scalar valued. At present, matrix-valued inputs should 
    only be used with equality constraints
    symvartable: List of the independent variables in the sosprogram
    '''
    
    if not decvartable:  # checks if decvartable is an empty list
        decvarnum = 0
    else:
        decvarnum = len(decvartable)

    if len(str(varmat))-2 == 2:
        vartable = symvartable
    else:
        vartable  = symvartable + varmat  # If they are sympy, + means concatenate.


    # If expr is a scalar, then FPexpr.shape is an empty tuple.
    # Thus, we need to check if expr is a Matrix.
    if isinstance(expr, sympy.MatrixBase):
        FPexpr = Matrix(expand(expr))
        dimp = FPexpr.shape[1]

        for i in range(FPexpr.shape[0]):
            for j in range(dimp):
                FPexpr[i,j] = collect(FPexpr[i,j], vartable)
                FPexpr[i,j] = together(FPexpr[i,j])         # Following are used to gaurantee the accuracy of the coefficients.
                FPexpr[i,j] = cancel(FPexpr[i,j])
                FPexpr[i,j] = poly_approx_zero(FPexpr[i,j])
                FPexpr[i,j] = cancel(FPexpr[i,j])

        # if decvartable is not empty.
        if decvartable:
            Zfull = np.array([])
            for i in range(dimp):
                for j in range(i, dimp):
                    p = poly(FPexpr[i, j], vartable)
                    coefmonmatr = [[*monomial] for monomial in p.monoms()]
                    Z = np.array(coefmonmatr)
                    Zfull = sortNoRepeat(Zfull, Z)

            Z = csr_matrix(Zfull)
            nmon, nvar = Z.shape
            coeffnts = lil_matrix((nmon * dimp, dimp))
            coeffnts_decvar_dict = {} # use this

            for i in range(dimp):
                for j in range(i, dimp):
                    p1 = poly(FPexpr[i, j], vartable)
                    coefmon = list(p1.coeffs())
                    # First I get all coefficients and monomials of the polynomial, 
                    # then I substitute all the decision variables with 0
                    coeff_bef_sub = p1.coeffs()
                    coeff_aft_sub = [cf.subs(dict(zip(decvartable, np.zeros(len(decvartable))))) for cf in coeff_bef_sub]
                    coefmonmatr = [[coeff, *monom] for coeff, monom in zip(coeff_aft_sub, p1.monoms())]

                    for k in range(len(coefmonmatr)):
                        s_ijk = coefmonmatr[k][0]

                        s_ijk_decvar = coefmon[k]

                        mon_k = coefmonmatr[k][1:]
                        ind_k = np.argmax(np.sum((Zfull == np.kron(np.ones((nmon,1)),mon_k)), axis=1))
                        coeffnts[ind_k*dimp+i, j] = s_ijk
                        coeffnts[ind_k*dimp+j, i] = s_ijk

                        # coeffnts_decvar is a sparse matrix so it cannot store Sympy values
                        # we can try to use dictionaly to store the values
                        coeffnts_decvar_dict[(ind_k*dimp+i, j)] = s_ijk_decvar
                        coeffnts_decvar_dict[(ind_k*dimp+j, i)] = s_ijk_decvar


        # if decvartable is empty.
        else: 
            Zfull = np.array([])
            for i in range(dimp):
                for j in range(i, dimp):
                    p = poly(FPexpr[i, j], vartable)
                    coefmon = [[*monomial] for monomial in p.monoms()]
                    Z = np.array(coefmon)
                    Zfull = sortNoRepeat(Zfull, Z)

            Z = csr_matrix(Zfull)
            nmon, nvar = Z.shape
            coeffnts = lil_matrix((nmon * dimp, dimp))

            for i in range(dimp):
                for j in range(i, dimp):
                    p = poly(FPexpr[i, j], vartable)
                    coefmon = [[coeff, *monom] for coeff, monom in zip(p.coeffs(), p.monoms())]

                    for k in range(len(coefmon)):
                        s_ijk = coefmon[k][0]
                        mon_k = coefmon[k][1:]
                        ind_k = np.argmax(np.sum((Zfull == np.kron(np.ones((nmon,1)),mon_k)), axis=1))
                        coeffnts[(ind_k)*dimp+i, j] = s_ijk
                        coeffnts[(ind_k)*dimp+j, i] = s_ijk


        
    # If expr is a scalar, then FPexpr.shape is an empty tuple. we cannot do FPexpr.shape[1].
    else:
        dimp = 1
        FPexpr = expand(expr)
        FPexpr = collect(FPexpr, vartable)
        FPexpr = together(FPexpr)
        FPexpr = cancel(FPexpr)
        FPexpr = poly_approx_zero(FPexpr)
        FPexpr = cancel(FPexpr)


        # if decvartable is not empty.
        if decvartable:
            Zfull = np.array([])
            p = poly(FPexpr, vartable)
            coefmonmatr = [[*monomial] for monomial in p.monoms()]
            Z = np.array(coefmonmatr)
            Zfull = sortNoRepeat(Zfull, Z)

            Z = csr_matrix(Zfull)
            nmon, nvar = Z.shape
            coeffnts = lil_matrix((nmon * dimp, dimp))
            coeffnts_decvar_dict = {} # use dictionary instead

            p1 = poly(FPexpr, vartable)
            coefmon = list(p1.coeffs())
            # First I get all coefficients and monomials of the polynomial, 
            # then I substitute all the decision variables with 0
            coeff_bef_sub = p1.coeffs()
            coeff_aft_sub = [cf.subs(dict(zip(decvartable, np.zeros(len(decvartable))))) for cf in coeff_bef_sub]
            coefmonmatr = [[coeff, *monom] for coeff, monom in zip(coeff_aft_sub, p1.monoms())]

            for k in range(len(coefmonmatr)):
                s_ijk = coefmonmatr[k][0]
                s_ijk_decvar = coefmon[k]

                mon_k = coefmonmatr[k][1:]
                ind_k = np.argmax(np.sum((Zfull == np.kron(np.ones((nmon,1)),mon_k)), axis=1))
                coeffnts[ind_k * dimp, 0] = s_ijk
                coeffnts_decvar_dict[(ind_k*dimp, 0)] = s_ijk_decvar

        # if decvartable is empty.
        else: 
            Zfull = np.array([])
            p = poly(FPexpr, vartable)
            coefmon = [[*monomial] for monomial in p.monoms()]
            Z = np.array(coefmon)
            Zfull = sortNoRepeat(Zfull, Z)

            Z = csr_matrix(Zfull)
            nmon, nvar = Z.shape
            coeffnts = lil_matrix((nmon * dimp, dimp))

            p = poly(FPexpr, vartable)
            coefmon = [[coeff, *monom] for coeff, monom in zip(p.coeffs(), p.monoms())]

            for k in range(len(coefmon)):
                s_ijk = coefmon[k][0]
                mon_k = coefmon[k][1:]
                ind_k = np.argmax(np.sum((Zfull == np.kron(np.ones((nmon,1)),mon_k)), axis=1))
                coeffnts[ind_k * dimp, 0] = s_ijk


    
    At = lil_matrix((decvarnum, Z.shape[0] * dimp ** 2))
    if decvartable:
        for k in range(Z.shape[0]):
            # Extract coefficients of decision variables from the dictionary
            Mivec = np.array([coeffnts_decvar_dict.get((i,j), 0) for i in range(k*dimp, (k+1)*dimp) for j in range(dimp)]).reshape(dimp**2,)
            jac_func = lambdify(decvartable, derive_by_array(Mivec, decvartable))
            # use * to unpack the list so that the function can take multiple arguments.
            At[:, k * dimp**2:(k+1) * dimp**2] = -jac_func(*([0] * len(decvartable)))   

    At = csr_matrix(At)
    b = csr_matrix(coeffnts)
    return At, b, Z


from scipy.sparse import csr_matrix, vstack, hstack, issparse


def spantiblkdiag(A1,A2):
    '''
    SPANTIBLKDIAG  Sparse anti block diagonal concatenation.

        A = spantiblkdiag(A1,A2)  

    Given two sparse matrices A1 and A2, this function
    constructs their anti block diagonal concatenation

        | 0  A1 |
    A = |       |
        | A2  0 |

    where A is a sparse matrix.
    '''

    A = vstack([hstack([csr_matrix((A1.shape[0],A2.shape[1])),A1]),hstack([A2,csr_matrix((A2.shape[0],A1.shape[1]))])])

    if not issparse(A):
        raise Exception('A is not sparse')
    
    return A





def mysymsubs(expr, old, new ,digit):
    '''
    MYSYMSUBS --- Fast symbolic substitution.

    NEWEXPR = mysymsubs(EXPR,OLD,NEW,DIGIT)

    EXPR is symbolic/string.
    OLD is symbolic.
    NEW is numeric (column vector).
    DIGIT is the number of digits.
    '''
    new = [round(item, digit) for item in new]
    replacements = {old[i]: new[i] for i in range(len(old))}
    new_expr = expr.subs(replacements)
    return new_expr


def sosgetsol(sos, V, digit=5):
    '''
    SOSGETSOL --- Get the solution from a solved SOS program 

    SOL = sosgetsol(SOSP,VAR,DIGIT) 

    SOL is the solution from a solved sum of squares program SOSP,
    obtained through substituting all the decision variables
    in VAR by the numerical values which are the solutions to
    the corresponding semidefinite program. 

    The third argument DIGIT (optional) will determine the 
    accuracy of SOL in terms of the number of digits. Default 
    value is 5.
    '''
    p = mysymsubs(V,sos.symdecvartable,sos.solinfo['RRx'][:len(sos.symdecvartable)],digit)
    return p




from mosek.fusion import *
import mosek
import numpy as np
import sys
import time

def sosmoseksolver(At,b,c,K,verbose=1):
    '''
    This function is to solve the optimization problem with MOSEK solver

    Output:
        x[list]: primal problem solution
        y[list]: dual problem solution
        info[dic]: information about the solution
    '''
    start_time = time.time()
    if 'q' in K:
        raise Exception('K.q and K.r are supposed to be empty')

    if 'f' not in K:
        K['f'] = []
    if 'l' not in K:
        K['l'] = []

    nf = int(np.sum(K['f']))
    nl = int(np.sum(K['l']))

    Ks2 = [int(item**2) for item in K['s']]
    Kindex = nf + nl + np.cumsum([0]+Ks2)
    RSPind = Ks2 + [nf]

    nsdpvar = len(K['s'])   # Number of Semidefinite Program Variables
    nvars = Kindex[0]       # Number of scalar variable

    # verbose = 1 default is 1, print out the information
    At_temp = At.toarray()
    c_temp = c.toarray()
    b_temp = b.toarray()

    with Model() as model:
        VAR_DIC = {}
        A_DIC = {}
        C_DIC = {}
        
        # Create Semidefinite variables
        for i in range(nsdpvar):   # Loop through Semidefinite Variable
            VAR_DIC[i] = model.variable(Domain.inPSDCone(K['s'][i]))
            A_DIC[i] = At_temp[Kindex[i]:Kindex[i+1]]
            C_DIC[i] = c_temp[Kindex[i]:Kindex[i+1]]
            
        # Create scalar variables, add to the end of VAR_DIC
        if nvars > 0:
            nidx = len(VAR_DIC)
            VAR_DIC[nidx] = model.variable(nvars)
            A_DIC[nidx] = At_temp[:nvars]
            C_DIC[nidx] = c_temp[:nvars]
            
        
        # Objective function and constraints
        if len(VAR_DIC)>1:   # More than 1 item in objective/constraints function
            OBJ_SUM = Expr.add(Expr.dot(C_DIC[0],VAR_DIC[0]),Expr.dot(C_DIC[1],VAR_DIC[1]))
            CON_SUM = Expr.add(Expr.mul(A_DIC[0].T,VAR_DIC[0].reshape(RSPind[0],1)),Expr.mul(A_DIC[1].T,VAR_DIC[1].reshape(RSPind[1],1)))
            for i in range(2,len(VAR_DIC)):
                OBJ_SUM = Expr.add(OBJ_SUM, Expr.dot(C_DIC[i],VAR_DIC[i]))
                CON_SUM = Expr.add(CON_SUM, Expr.mul(A_DIC[i].T,VAR_DIC[i].reshape(RSPind[i],1)))
        else:
            OBJ_SUM = Expr.dot(C_DIC[0], VAR_DIC[0])
            CON_SUM = Expr.mul(A_DIC[0].T, VAR_DIC[0].reshape(RSPind[0],1))
            
        # Objective function
        obj = model.objective(ObjectiveSense.Minimize, OBJ_SUM)
        
        # Constraints
        constr = model.constraint(Expr.sub(CON_SUM, b_temp), Domain.equalsTo(0.0))
        
        if verbose == 1:
            model.setLogHandler(sys.stdout)
        # Solve the model
        
        model.solve()
        end_time = time.time()

        # Get status information about the solution
        feasbl = model.getProblemStatus()
        if feasbl != ProblemStatus.PrimalAndDualFeasible:
            sol = []
            y = np.array([])
        else:
            sol = []
            y = constr.dual()
            if nvars > 0:
                sol.append(VAR_DIC[nidx].level())
                nidx = 1
            else:
                nidx = 0
            for i in range(len(VAR_DIC)-nidx):
                sol.append(VAR_DIC[i].level())

        task = model.getTask()
        info = {}
        info['cpusec'] = round(end_time-start_time,5)
        info['iter'] = task.getintinf(mosek.iinfitem.intpnt_iter)
        info['feasratio'] = round(task.getdouinf(mosek.dinfitem.intpnt_opt_status),5)
        if feasbl == ProblemStatus.PrimalInfeasible:
            info['pinf'] = 1
        else:
            info['pinf'] = 0
        if feasbl == ProblemStatus.DualInfeasible:
            info['dinf'] = 1
        else:
            info['dinf'] = 0
        #info['numerr'] = 0

        # convert sol and y to list
        x = []
        for item in sol:
            x.extend(item.flatten().tolist())
        y = y.flatten().tolist()

    return x, y, info